#!/usr/bin/perl -w

##############################################################################
# Copyright 2006 Sergei Steshenko
# (
#  sergei_steshenko _AT_ pop3.ru
#  sergei_steshenko _AT_ softhome.net
# ).
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# If someone modifies the program to run with LGPL-compatible FFT, the
# he/she can also release the derived program under the terms of LGPL, as
# published by Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
##############################################################################

require File::Basename;
require File::Path;
require File::Find;
require Cwd;
# I do not want the above to export names, that's why 'require',
# not 'use'

use strict;

my $_tool_name;
my $_tool_path;
($_tool_name, $_tool_path, undef) = &File::Basename::fileparse($0);


my $_info_marker = "$_tool_name :INFO:";
my $_warning_marker = "$_tool_name :WARNING:";
my $_error_marker = "$_tool_name :ERROR:";

my $_message="# $0 will be truncated to $_tool_name in error/warning/info messages #";
warn '#' x length($_message),"\n";
warn $_message,"\n";
warn '#' x length($_message),"\n";

warn "Main program of AppsFromScratch - distributed under the terms of GPL. Copyright (c) Sergei Steshenko, 2006\n\n";


require "$_tool_path/../include/perl/general/ArrayHashDumper.pm";

$ArrayHashDumper::dump_indent_offset=$ArrayHashDumper::dump_indent_offset=0;
$ArrayHashDumper::dump_indent_factor=$ArrayHashDumper::dump_indent_factor=2;
$ArrayHashDumper::tool_name=$ArrayHashDumper::tool_name=$_tool_name;
$ArrayHashDumper::error_marker=$ArrayHashDumper::error_marker=':ERROR:';

my $_failed_banner_sub = require "$_tool_path/../include/perl/general/failed_banner.prl";
my $_system_wrapper_sub = require "$_tool_path/../include/perl/general/system_wrapper.prl";
my $_backticks_wrapper_sub = require "$_tool_path/../include/perl/general/backticks_wrapper.prl";
my $_shuffle_download_sub = require "$_tool_path/../include/perl/general/shuffle_download.prl";
my $_rerequire_sub = require "$_tool_path/../include/perl/general/rerequire.prl";
my $_gen_versioned_target_name_sub = require "$_tool_path/../include/perl/project_specific/gen_versioned_target_name.prl";

my $_layer_build_sub;
my $_target_build_sub;
my $_build_data_sub;

my $_cwd = Cwd::cwd('.');

my %_args;
my %_targets_to_build;
my %_targets_to_exclude;

&deal_with_command_line();
warn "\n";

unless(exists $_args{build_prefix})
  {
  $_args{build_prefix} = "$_cwd/build";
  warn "$_info_marker using default '$_args{build_prefix}' build prefix directory\n";
  }

unless(exists $_args{install_prefix})
  {
  $_args{install_prefix} = "$_cwd/install";
  warn "$_info_marker using default '$_args{install_prefix}' install prefix directory\n";
  }

unless(exists $_args{default_tarball_extension})
  {
  $_args{default_tarball_extension} = 'tar.bz2';
  warn "$_info_marker using default '$_args{default_tarball_extension}' tarball extension\n";
  }

&dump_env();

if(ref($_layer_build_sub) ne 'CODE')
  {
  warn "$_error_marker \$_layer_build_sub is not a CODE reference\n";
  &{$_failed_banner_sub}($_tool_name);
  die;
  }

my $_top_level_build_data_hash_ref =
&{$_build_data_sub}
  (
  cmdline_args_hash_ref => \%_args,
  tool_name => $_tool_name,
  info_marker => $_info_marker,
  warning_marker => $_warning_marker,
  error_marker => $_error_marker,
  system_wrapper_sub => $_system_wrapper_sub,
  backticks_wrapper_sub => $_backticks_wrapper_sub,
  failed_banner_sub => $_failed_banner_sub,
  rerequire_sub => $_rerequire_sub,
  gen_versioned_target_name_sub => $_gen_versioned_target_name_sub,
  default_layer_build_sub => $_layer_build_sub,
  default_target_build_sub => $_target_build_sub
  );


if(ref($_top_level_build_data_hash_ref) ne 'HASH')
  {
  warn "$_error_marker \$_build_data_sub did not return a HASH reference\n";
  &{$_failed_banner_sub}($_tool_name);
  die;
  }

my $_targets_hash_ref;

if(exists ${$_top_level_build_data_hash_ref}{targets})
  {
  $_targets_hash_ref = ${$_top_level_build_data_hash_ref}{targets};
  }
else
  {
  warn "$_error_marker \${\$_top_level_build_data_hash_ref}{targets} does not exist\n";
  &{$_failed_banner_sub}($_tool_name);
  die;
  }

if(exists $_args{list_top_level_targets})
  {
  warn "$_info_marker list of available top level targets:\n";

  ###################################################
  foreach my $target(sort keys %{$_targets_hash_ref})
  ###################################################
    {
    warn $target, "\n";
    } # foreach my $target(sort keys %{$_targets_hash_ref})

  warn "$_info_marker END OF list of available top level targets:\n";
  exit(0);
  } # if(exists $_args{list_top_level_targets})

my $_recursion_level = 0;
my @_path;

my $result;

my %_layer_args =
  (
  cmdline_args_hash_ref => \%_args,
  recursion_level_scalar_ref =>\$_recursion_level,
  path_array_ref => \@_path,
  tool_name => $_tool_name,
  info_marker => $_info_marker,
  warning_marker => $_warning_marker,
  error_marker => $_error_marker,
  system_wrapper_sub => $_system_wrapper_sub,
  backticks_wrapper_sub => $_backticks_wrapper_sub,
  failed_banner_sub => $_failed_banner_sub,
  shuffle_download_sub => $_shuffle_download_sub,
  rerequire_sub => $_rerequire_sub,
  gen_versioned_target_name_sub => $_gen_versioned_target_name_sub,
  top_level_hash_ref => $_top_level_build_data_hash_ref,
  targets_hash_ref => $_targets_hash_ref,
  targets_to_build_hash_ref => \%_targets_to_build,
  targets_to_exclude_hash_ref => \%_targets_to_exclude,
  default_layer_build_sub => $_layer_build_sub,
  default_target_build_sub => $_target_build_sub
  );

if(exists ${$_top_level_build_data_hash_ref}{layer_build_sub})
  {
  if(ref(${$_top_level_build_data_hash_ref}{layer_build_sub}) ne 'CODE')
    {
    warn "$_error_marker value of \${\$_top_level_build_data_hash_ref}{layer_build_sub} key is not a code reference\n";
    &{$_failed_banner_sub}($_tool_name);
    die;
    }

  warn "$_info_marker calling supplied through build data layer build sub\n";

    (
    $result,
    undef,
    undef
    ) = &{${$_top_level_build_data_hash_ref}{layer_build_sub}}(%_layer_args);
  }
else
  {
  warn "$_info_marker calling supplied through $_tool_name layer build sub\n";

    (
    $result,
    undef,
    undef
    ) = &{$_layer_build_sub}(%_layer_args);
  }

if($result == 0)
  {
  $_message="+ $_tool_name SUCCESSFULLY COMPLETED +";
  warn '+' x length($_message),"\n";
  warn $_message,"\n";
  warn '+' x length($_message),"\n";

  exit(0);
  }
else
  {
  warn "$_error_marker build failed\n";
  &{$_failed_banner_sub}($_tool_name);
  die;
  }
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------


##########################
sub deal_with_command_line
##########################
  {
  my @file_options =
  qw(
    build_data_sub_file
    layer_build_sub_file
    target_build_sub_file
    );

  my @dir_options =
  qw(
    build_prefix
    install_prefix
    );

  my @command_options;
  my @flags =
  qw(
    list_top_level_targets
    make_like
    dont_cleanup_install_dirs
    );

  my @string_options =
  qw(
    targets_to_build
    targets_to_exclude
    default_tarball_extension
    extra_args
    );

  my @number_options;

  my @mandatory_options;

  my $usage = <<EOD
USAGE:

$0 \\
[-build_prefix <build_prefix_dir>] \\
[-install_prefix <install_prefix_dir>] \\
[-targets_to_build <colon_separated_list_of_targets>] \\
[-targets_to_exclude <colon_separated_list_of_targets>] \\
[-build_data_sub_file <build_data_sub_perl_file>] \\
[-layer_build_sub_file <layer_build_sub_file_perl_file>] \\
[-target_build_sub_file <target_build_sub_file_perl_file>] \\
[-list_top_level_targets] \\
[-make_like] \\
[-dont_cleanup_install_dirs] \\
[-default_tarball_extension <default_tarball_extension>] \\
[-extra_args <args_for_build_data_sub>] \\


General.
--------

The script by itself contains almost no end user useful functionality - all the
build job is done by build subroutines which might be provided by user
(instead of the default ones).

So, the functionality description below is relevant for the default build
subroutines.

Messages produced by the script and the default build subroutined are printed
to STDERR.


Defaults, target locations.
---------------------------

If invoked without argumets, the script uses default

$_tool_path/../include/perl/project_specific/default_target_build_sub.prl
$_tool_path/../include/perl/project_specific/default_layer_build_sub.prl

build subroutines and builds targets described in the hash (reference) returned
by default

$_tool_path/../include/perl/project_specific/default_build_data_sub.prl

subroutine.

Default build prefix directory is './build', i.e. targets are
downloaded + unpackaged into and built under

./build/<target_basename>-<major_ver>.<minor_ver>.<micro_ver>

directories, for example:

build/atk-1.10.3.tar.gz - downloaded from internet source tarball;
build/atk-1.10.3/       - the above unpackaged tarball, aslo build subdir.

Similarly, default install prefix is './install', i.e. targets are installed
as

./install/<target_basename>-<major_ver>.<minor_ver>.<micro_ver>

directories, for example:

install/atk-1.10.3/
.


A non-default build prefix is specified as

-build_prefix <build_prefix_dir>

- if so, the targets are downloaded as

<build_prefix_dir>/<target_basename>-<major_ver>.<minor_ver>.<micro_ver>.tar.gz

and build directories are

<build_prefix_dir>/<target_basename>-<major_ver>.<minor_ver>.<micro_ver>/
.


Likewise, a non-default install prefix is specified as

-install_prefix <install_prefix_dir>

- if so, the targets are installed as

<install_prefix_dir>/<target_basename>-<major_ver>.<minor_ver>.<micro_ver>/
.

Per target install directories can be specified in the build data subroutine -
see README file for details.


The default build data subroutine can be replaced by user supplied one using

-build_data_sub_file <build_data_sub_perl_file>

option.

Read detailed documentation on how to write build data subroutines.


Likewise, the default layer and target build subroutines can be replaced by
user supplied ones using

-layer_build_sub_file <layer_build_sub_file_perl_file> \\
-target_build_sub_file <target_build_sub_file_perl_file> \\

options.

Analyze in detail code of the default

$_tool_path/../include/perl/project_specific/default_layer_build_sub.prl
$_tool_path/../include/perl/project_specific/default_target_build_sub.prl

subroutines first.


Targets to be built and/or to be excluded.
------------------------------------------

If neither '-targets_to_build ...' nor '-targets_to_exclude ...' is specified,
all the targets (and their dependencies) specified at recursion level zero
of the hash (reference) returned by <build_data_sub_perl_file> are built.

If '-targets_to_build <colon_separated_list_of_targets>' is specified, only
targets found in <colon_separated_list_of_targets> are built.

If '-targets_to_exclude <colon_separated_list_of_targets>' is specified,
the targets found in <colon_separated_list_of_targets> are excluded from the
internal list of targets to be built - again, it is related to recursion level
zero, dependencies are built as necessary.

If both '-targets_to_build ...' and '-targets_to_exclude ...' are specified,
the latter targets are excluded from the former.


Listing targets.
----------------

if '-list_top_level_targets' is specified, all the targets specified at
recursion level zero of the hash (reference) returned by
<build_data_sub_perl_file> will be listed one per line and the script
will exit not building anything.


Make-like behavior, recursion.
------------------------------

Target dependencies are described in the hash (reference) exported by build data
subroutine (see above '-build_data_sub_file ...'). The hash is hierarchical, the
dependencies are the hierarchy.

The hash is traversed depth first, so this ensures that first all dependencies
are built and only then the target itself is built.

At the very end of build + install process (after 'make install') a zero-size
file is created to mark that the target has been built and isntalled.

The file is

<targete_install_dir>/___ALREADY_BUILT___.<target_basename>-<major_ver>.<minor_ver>.<micro_ver>

, for example:

install/atk-1.10.3/___ALREADY_BUILT___.atk-1.10.3
.

Let\'s call the file "currency file".

If the script is run again, after targets are built, currency file is used
to determine whether a target has been built.

If currency file does not exist, the target will be rebuilt as if has never
been built.

If currency file exists, it is depending on presence of '-make_like' on
command line how it is used.

If '-make_like' is not present, then at the moment it detected currency file
exists recursion stops and a message like

build.pl :INFO: doing nothing to build 'Glib' target since it has already been built and installed

is issued.

If '-make_like' is present, recursion goes all the way down to leaf targets;
at each level target (its currency file) is checked against the newest
currency file of targets it depends on. "Checked against" means modification
times are compared.

If a target\'s currency file modification time is smaller than the one of the
newest target this target depends on, the target is rebuilt.

Rebuilding the target means creation of new currency file, so in such a manner
rebuilding lower level targets triggers rebuilding of higher level targets.

In this respect presence of '-make_like' on commane line makes the script
behave in a manner similar to 'make' utility.


(Not) cleaning up install directories.
--------------------------------------

By default if a target is rebuilt, its install directory is cleaned up
using 'rm -rf', i.e. it is completely completely removed.

This is fine if each target is installed into a separate directory (also
the default).

It is possible to specify per target install prefix in build data subroutine,
and thus it is possible to install all targets under the same directory
- if all targets have te same install prefix.

If all targets are installed in the same directory, cleaning up one target
cleans up them all.

It is possible to disable all targets cleanup using

-dont_cleanup_install_dirs

command line option.


Tarball extension.
------------------

By default target tarballs are supposed to be named <versioned_target>.tar.bz2.

The 'tar.bz2' default can be changed globally using

-default_tarball_extension <default_tarball_extension>

option and/or on per target basis - see README file for details.


Extra arguments.
----------------

Specifying '-extra_args <args_for_build_data_sub>' on command line user may
supply extra arguments to both build and build data subroutines (see above
-build_data_sub_file, -layer_build_sub_file).

If more than one word is supposed to be passed in <args_for_build_data_sub>,
it will have to be quoted on command line.

Actually, command line arguments are stored in a hash, and reference to the hash
is passed to both, build and build data subroutines.

Should the first item in <args_for_build_data_sub> begin with '-' sign, it
should be prepended with space because of the way command line parser works in
this script. This is intentional.

This help message is printed to STDERR and can be redirected to a file
using

$0 -h 2>built_help_message.txt

command in 'sh'-like shells or

$0 -h >&! built_help_message.txt

in 'csh'-like shells.

EOD
  ;

  my $syntax_error;

  #########################################
  while(defined(my $argv_item=shift @ARGV))
  #########################################
    {
    if($argv_item=~m/^--?(\w+)$/)
      {
      my $option = $1;

      if(grep($_ eq $option, @file_options))
	{
	$_args{$option} = shift @ARGV;

	if(
           (defined($_args{$option}) and ($_args{$option}=~m/^-/))
	or not defined($_args{$option})
          )
          {
          warn "$_error_marker '-$option' should be followed by file name";
          $syntax_error = 1;
          last;
          }
	}
      elsif(grep($_ eq $option, @dir_options))
	{
	$_args{$option} = shift @ARGV;

	if(
           (defined($_args{$option}) and ($_args{$option}=~m/^-/))
	or not defined($_args{$option})
          )
          {
          warn "$_error_marker '-$option' should be followed by directory name";
          $syntax_error = 1;
          last;
          }
	}
      elsif(grep($_ eq $option, @string_options))
	{
	$_args{$option} = shift @ARGV;

	if(
           (defined($_args{$option}) and ($_args{$option}=~m/^-/))
	or not defined($_args{$option})
          )
          {
          warn "$_error_marker '-$option' should be followed by string";
          $syntax_error = 1;
          last;
          }
	}
      elsif(grep($_ eq $option, @number_options))
	{
	$_args{$option} = shift @ARGV;

	if(
	  not defined($_args{$option})
        or not ($_args{$option} =~ m/^\d+$/)
	  )
          {
          warn "$_error_marker '-$option' should be followed by unsigned integer number";
          $syntax_error = 1;
          last;
          }
	}
      elsif(grep($_ eq $option, @command_options))
	{
	$_args{$option} = shift @ARGV;

	if(
           (defined($_args{$option}) and ($_args{$option}=~m/^-/))
	or not defined($_args{$option})
          )
          {
          warn "$_error_marker '-$option' should be followed by command";
          $syntax_error = 1;
          last;
          }
	}
      elsif(grep($_ eq $option, @flags))
	{
	$_args{$option} = undef;
	}
      else
	{
	unshift @ARGV, $argv_item;
	warn "$_error_marker unrecognized '$argv_item' command line option";
	warn "$_error_marker unparsed remainder of command line: @ARGV";
	$syntax_error = 1;
	last;
	}
      }
    else
      {
      unshift @ARGV, $argv_item;
      last;
      }
    } # while(defined(my $argv_item=shift @ARGV))


  my $mandatory_option_is_missing;
  map
    {
    unless(exists $_args{$_})
      {
      $mandatory_option_is_missing = 1;
      warn "$_error_marker '-$_' mandatory command line option is missing\n";
      }
    } @mandatory_options;

  if(scalar(@ARGV) != 0)
    {
    warn "$_error_marker unparsed remainder of command line: @ARGV";
    $syntax_error = 1;
    }

  if($mandatory_option_is_missing or $syntax_error)
    {
    die $usage,' died';
    }

  warn "$_info_marker processed command line arguments in internal Perl form:\n";
  &ArrayHashDumper::dump_array_hash
    (
    \%_args,  # $ref, 
    '_args',  # $name,
    \*STDERR, # $fh,
    );

  if(exists $_args{build_data_sub_file})
    {
    unless(-e $_args{build_data_sub_file})
      {
      warn "$_error_marker '$_args{build_data_sub_file}' file does not exist\n";
      &{$_failed_banner_sub}($_tool_name);
      die;
      }

    warn "$_info_marker going to 'require \"$_args{build_data_sub_file}\"'\n";
    $_build_data_sub = require $_args{build_data_sub_file};
    }
  else
    {
    $_build_data_sub = $_build_data_sub = require "$_tool_path/../include/perl/project_specific/default_build_data_sub.prl";
    }

  if(ref($_build_data_sub) ne 'CODE')
    {
    warn "$_error_marker $_build_data_sub is not a CODE reference\n";
    &{$_failed_banner_sub}($_tool_name);
    die;
    }


  if(exists $_args{layer_build_sub_file})
    {
    unless(-e $_args{layer_build_sub_file})
      {
      warn "$_error_marker '$_args{build_data_sub_file}' file does not exist\n";
      &{$_failed_banner_sub}($_tool_name);
      die;
      }

    warn "$_info_marker going to 'require \"$_args{layer_build_sub_file}\"'\n";
    $_layer_build_sub = require $_args{layer_build_sub_file};
    }
  else
    {
    $_layer_build_sub = require "$_tool_path/../include/perl/project_specific/default_layer_build_sub.prl";
    }

  if(exists $_args{target_build_sub_file})
    {
    unless(-e $_args{target_build_sub_file})
      {
      warn "$_error_marker '$_args{build_data_sub_file}' file does not exist\n";
      &{$_failed_banner_sub}($_tool_name);
      die;
      }

    warn "$_info_marker going to 'require \"$_args{target_build_sub_file}\"'\n";
    $_target_build_sub = require $_args{target_build_sub_file};
    }
  else
    {
    $_target_build_sub = require "$_tool_path/../include/perl/project_specific/default_target_build_sub.prl";
    }

  if(exists $_args{targets_to_build})
    {
    map {$_targets_to_build{$_} = ''} split(':', $_args{targets_to_build});
    }

  if(exists $_args{targets_to_exclude})
    {
    map {$_targets_to_exclude{$_} = ''} split(':', $_args{targets_to_exclude});
    }
  } # sub deal_with_command_line
  #-----------------------------


############
sub dump_env
############
  {
  warn "$_info_marker dumping environment - please examine it carefully\n\n";

  foreach my $env_var(sort keys %ENV)
    {
    warn "$env_var=$ENV{$env_var}\n";
    } # foreach my $env_var(sort keys %ENV)

  warn "\n$_info_marker DONE dumping environment\n\n";
  } # sub dump_env
  #---------------
