###############################################################################
# Copyright 2002 Sergei Steshenko
# (
#  sergei_steshenko _AT_ pop3.ru
#  sergei_steshenko _AT_ softhome.net
# ).
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the Lesser/Library GNU General Public License (LGPL) as
# published by the Free Software Foundation; either version 2 of the License,
# or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# If someone modifies the program to run with LGPL-compatible FFT, the
# he/she can also release the derived program under the terms of LGPL, as
# published by Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
###############################################################################

use FileHandle;
use Sys::Hostname;
use B::Deparse;

        ###################
package ArrayHashCodeDumper;
        ###################


use strict;

our
(
 $hash_key_sorting_and_access_code_ref,
 $dump_indent_factor,
 $dump_indent_offset,
 $tool_name,
 $error_marker,
 $ref_to_itself_basename,
 $access_code_marker,
 $failure_return_code,
 $success_return_code,
 $uniquify_references,
);

use integer;


my $recursion_level=0;


$tool_name=(reverse split('/',$0))[0];
$error_marker=':E:';
$ref_to_itself_basename='___ref_to_itself___';
$access_code_marker='access code: ';
$failure_return_code=1;
$success_return_code=0;


# PERL_BEGIN
#############################
# written by Sergei Steshenko
#############################

#  Copyright 2001 Sergei Steshenko 
#  (
#   sergei.steshenko@usa.net
#   sergei_steshenko@my-deja.com
#  ).
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation; either version 2 of the License, or (at your option)
#  any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
#  for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


my %recursion_levels_to_sorted_keys;

#####################################
$hash_key_sorting_and_access_code_ref=
#####################################
sub{
   my ($hash_ref)=@_;

   if(!exists $recursion_levels_to_sorted_keys{$recursion_level})
     {
     @{$recursion_levels_to_sorted_keys{$recursion_level}}=
     sort keys %{$hash_ref};
     }

   my $key=shift @{$recursion_levels_to_sorted_keys{$recursion_level}};

   if(defined $key)
     {
     return ($key,${$hash_ref}{$key},); # the returned list
     }
   else
     {
     delete $recursion_levels_to_sorted_keys{$recursion_level};
     return undef;
     }
   };
#-------------------------------------------
# PERL_END

my %_path;
# the above hash will be used to detect looping, i.e. if a data structure
# references itself

my @_refs_to_itself;
my %_refs_to_itself_to_numbers;
# the above array and hash will be used to store references which reference
# themselves

my $_ref_to_itself_number;

my %_unique_refs_to_numbers;
my $_unique_ref_number;

    ####################
sub study_data_structure
    ####################
{
my ($ref)=@_;# warn "study_data_structur called; \$recursion_level=$recursion_level";

my $stringified_ref="$ref";

if(exists $_path{$stringified_ref})
  {# warn "loop fond: \$stringified_ref=$stringified_ref";
  unless(exists $_refs_to_itself_to_numbers{$stringified_ref})
    {
    push @_refs_to_itself,$ref;
    $_refs_to_itself_to_numbers{$stringified_ref}=
    $_ref_to_itself_number++;
    }

  return;
  }
else
  {
  $_path{$stringified_ref}=$recursion_level;

  if($uniquify_references and ($recursion_level > 0))
    {
    unless(exists $_unique_refs_to_numbers{$stringified_ref})
      {
      unless(defined $_unique_ref_number)
        {
        $_unique_ref_number=0;
        }

      $_unique_refs_to_numbers{$stringified_ref}=$_unique_ref_number++;
      }
    }
  }

#+++++++++++++++
if(not ref $ref)
#+++++++++++++++
  {
  die "$tool_name $error_marker \$ref=$ref should be a reference, died";
  }
#++++++++++++++++++++++++
elsif(ref $ref eq 'CODE')
#++++++++++++++++++++++++
  {
  delete $_path{$stringified_ref};
  return # do nothing
  }
#++++++++++++++++++++++++
elsif(ref $ref eq 'HASH')
#++++++++++++++++++++++++
  {
  ####################################
  while(my ($key,$value)=each %{$ref})
  ####################################
    {
    if(ref $value)
      {
      #################
      # recursive call:
      #################
      $recursion_level++;
      &study_data_structure($value);
      $recursion_level--;
      }
    } # while(($key,undef)=each %{$ref})
  }
#+++++++++++++++++++++++++
elsif(ref $ref eq 'ARRAY')
#+++++++++++++++++++++++++
  {
  ############################
  foreach my $element(@{$ref})
  ############################
    {
    if(ref $element)
      {
      #################
      # recursive call:
      #################
      $recursion_level++;
      &study_data_structure($element);
      $recursion_level--;
      }
    } # foreach my $element(@{$ref})
  }
else
  {
  die "$tool_name $error_marker unsupported reference type \$ref=$ref, died";
  }

delete $_path{$stringified_ref};
} # sub study_data_structure
#---------------------------

my $_dont_study;
my $_deparse_instance=new B::Deparse "-si2";
my $_name;
my $_stringified_top_level_ref;
my @_access_strings;

    ####################
sub dump_array_hash_code
    ####################
{
my
(
 $ref,
 $scope_for_name, # can be 'my' or 'local'
 $name,
 $scope_for_self, # can be 'my' or 'local'
 $name_for_self,  # no '$' here !
 $fh,
)=@_; # warn "\$ref=$ref \$recursion_level=$recursion_level";


if((0 == $recursion_level) and (not $_dont_study))
  {
  if(
     (defined $name)
 and ($name ne '')
 and (defined $name_for_self)
 and ($name_for_self ne '')
    )
    {
    die "$tool_name $error_marker both '$name' and '$name_for_self' were specified, they are mutually exclusive";
    }

  if((defined $name) and ($name ne ''))
    {
    $_access_strings[0]=$name
    }
  elsif((defined $name_for_self) and ($name_for_self ne ''))
    {
    $_access_strings[0]="\$$name_for_self"
    }
  else
    {
    $_access_strings[0]='<actual_name>';
    }

  &study_data_structure(do{$_stringified_top_level_ref="$ref";$ref});

  $_dont_study=1;

  map
    {
    if(
       (defined $name)
   and ($name ne '')
   and ("$_" eq $_stringified_top_level_ref)
      )
      {
      if((defined $scope_for_name) and ($scope_for_name ne ''))
        {
        my $percent_or_at=
        ref($ref) eq 'ARRAY' ? '@' : ref($ref) eq 'HASH' ? '%' : undef
        ;
        if(defined $percent_or_at)
          {
          print $fh "$scope_for_name $percent_or_at$name;\n";
          }
        }
      else
        {
        warn "$tool_name $error_marker data structure to be dumped references itself and is to be dumped with \$name=$name name - this requires \$scope_for_name subroutine argument to be set; nothing will be dumped, exiting with \$failure_return_code=$failure_return_code";

        $_ref_to_itself_number=0;
        undef($_dont_study);
        undef(@_refs_to_itself);
        undef(%_refs_to_itself_to_numbers);
        undef($_name);
        undef($_stringified_top_level_ref);
        undef(@_access_strings);

        return $failure_return_code;
        }
      }
    elsif
      (
       (defined $name_for_self)
   and ($name_for_self ne '')
   and ("$_" eq $_stringified_top_level_ref)
      )
      {
      if((defined $scope_for_self) and ($scope_for_self ne ''))
        {
        print $fh "$scope_for_self \$$name_for_self;\n";
        }
      else
        {
        warn "$tool_name $error_marker data structure to be dumped references itself and is to be dumped with \$name_for_self=$name_for_self name for anonymous object to be exported - this requires \$$scope_for_self subroutine argument to be set; nothing will be dumped, exiting with \$failure_return_code=$failure_return_code";

        $_ref_to_itself_number=0;
        undef($_dont_study);
        undef(@_refs_to_itself);
        undef(%_refs_to_itself_to_numbers);
        undef($_name);
        undef($_stringified_top_level_ref);
        undef(@_access_strings);

        return $failure_return_code;
        }
      }
    else
      {
      print $fh "my \$$ref_to_itself_basename$_refs_to_itself_to_numbers{$_};\n"
      }
    } @_refs_to_itself;

  ########################################################
  while(defined(my $ref_to_itself=shift @_refs_to_itself))
  ########################################################
    {
    my $stringified_ref_to_itself="$ref_to_itself";

    unless($stringified_ref_to_itself eq $_stringified_top_level_ref)
      {
      print $fh "\$$ref_to_itself_basename$_refs_to_itself_to_numbers{$stringified_ref_to_itself}=";

      #################
      # recursive call:
      #################
      &dump_array_hash_code
      (
       $ref_to_itself,  # $ref,
       $scope_for_name, # $scope_for_name, # can be 'my' or 'local'
       undef,           # $name,
       $scope_for_self, # can be 'my' or 'local'
       $name_for_self,  # no '$' here !
       $fh,             # $fh,
      );

      print $fh "\n";
      }
    } # while(defined(my $ref_to_itself=shift @_refs_to_itself))

  if(exists $_refs_to_itself_to_numbers{$_stringified_top_level_ref})
    {
    if((defined $name_for_self) and ($name_for_self ne ''))
      {
      print $fh "\$$name_for_self=";
      }
    elsif(not (defined $name) or ($name eq ''))
      {
      my $ref_to_itself_number=
      $_refs_to_itself_to_numbers{$_stringified_top_level_ref};
      print $fh "\$$ref_to_itself_basename$ref_to_itself_number=";
      }
    }

  undef($_dont_study);
  } # if(0 == $recursion_level and not $_dont_study)

if(not ref $ref)
  {
  die "$tool_name $error_marker \$ref=$ref should be a reference, died";
  }
elsif(($recursion_level) > 0 and (exists $_refs_to_itself_to_numbers{"$ref"}))
  {
  if("$ref" eq $_stringified_top_level_ref)
    # if reference to itself is reference to top level reference
    {
    if((defined $_name) and ($_name ne ''))
      {
      print $fh "\\$_name"; # top level name is used
      }
    elsif((defined $name_for_self) and ($name_for_self ne ''))
      {
      print $fh "\$$name_for_self"; # name for self is used
      }
    else
      {
      my $ref_to_itself_number=$_refs_to_itself_to_numbers{"$ref"};
      print $fh "\$$ref_to_itself_basename$ref_to_itself_number";
      }
    }
  else
    {
    my $ref_to_itself_number=$_refs_to_itself_to_numbers{"$ref"};
    print $fh "\$$ref_to_itself_basename$ref_to_itself_number";
    }
  }
else
  {
  if(0 == $recursion_level)
    {
    if((defined $name) and ($name ne ''))
      {
      $_access_strings[0]=$name;
      }
    elsif((defined $name_for_self) and ($name_for_self ne ''))
      {
      $_access_strings[0]="\$$name_for_self";
      }
    else
      {
      $_access_strings[0]='<actual_name>';
      }
    }

  my $spacer=
  ' ' x ($dump_indent_factor * $recursion_level + $dump_indent_offset);

  my $anonymous_style_printing=
  (($recursion_level != 0) or (not defined $name) or ($name eq ''));

  #++++++++++++++++++++++
  if(ref $ref eq 'ARRAY')
  #++++++++++++++++++++++
    {
    print $fh
    $anonymous_style_printing
    ? "\n$spacer\["
    : do{
        $_name="%$name";
        ($spacer,"$_name=\n$spacer(");
        }
    ;
    my $comma='';
    my $number_of_elements=scalar @{$ref};

    my $decremented_el_number;
    my $el_number;
    ###############################################################
    for($el_number=0;$el_number < $number_of_elements;$el_number++)
    ###############################################################
      {
      $decremented_el_number=$el_number-1;

      my $element=${$ref}[$el_number];

      print $fh
      $comma,
      $decremented_el_number >= 0
      ? (
         ' # '.$decremented_el_number.' #'
        ." $access_code_marker\$\{$_access_strings[$recursion_level]\}[$decremented_el_number]"
        )
      : '';

      if(ref $element)
        {
	#################
	# recursive call:
	#################
        $_access_strings[$recursion_level + 1]=
        "\${$_access_strings[$recursion_level]}[$el_number]";

	$recursion_level++;
        &dump_array_hash_code
	(
	 $element,        # $ref, 
         $scope_for_name, # $scope_for_name, # can be 'my' or 'local'
	 undef,           # $name,
         $scope_for_self, # can be 'my' or 'local'
         $name_for_self,  # no '$' here !
	 $fh,             # $fh,
	);

        $recursion_level--;
	}
      else
        {
	print $fh "\n",$spacer;

        if(defined $element)
	  {
	  $element=~s/\\/\\\\/gs;
	  $element=~s/\"/\\"/gs;
	  $element=~s/\$/\\\$/gs;
	  $element=~s/\@/\\@/gs;

	  print $fh '"'.$element.'"';
	  }
        else
	  {
	  print $fh 'undef';
	  }
	}

      $comma=',';
      } # for($el_number=0;$el_number < $number_of_elements;$el_number++)

    if(defined $decremented_el_number)
      {
      $decremented_el_number++;
      print $fh '  # '.$decremented_el_number.' #';
      print $fh " $access_code_marker\$\{$_access_strings[$recursion_level]\}[$decremented_el_number]"
      }

    print $fh "\n",$spacer;

    print $fh $anonymous_style_printing ? ']' : ')';
    }
  #++++++++++++++++++++++++
  elsif(ref $ref eq 'HASH')
  #++++++++++++++++++++++++
    {
    print $fh
    $anonymous_style_printing
    ? "\n$spacer\{"
    : do{
        $_name="%$name";
        ($spacer,"$_name=\n$spacer(");
        }
    ;
    my $comma='';
    my $key;
    my $previous_key;
    my $value;
    ##################################################################
    while
      (
       do{
	 ($key,$value)=&{$hash_key_sorting_and_access_code_ref}($ref);
	 defined $key;
         }
      )
    ##################################################################
      {
      print $fh $comma;

      if(defined $previous_key)
        {
        print $fh " # $access_code_marker\${$_access_strings[$recursion_level]}{$previous_key}"
        }

      $key=~s/\\/\\\\/gs;
      $key=~s/\"/\\"/gs;
      $key=~s/\$/\\\$/gs;
      $key=~s/\@/\\@/gs;

      if(ref $value)
        {
        print $fh "\n",$spacer,'"',$key,'" => ';

	#################
	# recursive call:
	#################
        $_access_strings[$recursion_level + 1]=
        "\${$_access_strings[$recursion_level]}{$key}";

	$recursion_level++;

        &dump_array_hash_code
	(
	 $value,          # $ref,
         $scope_for_name, # $scope_for_name, # can be 'my' or 'local'
	 undef,           # $name,
         $scope_for_self, # can be 'my' or 'local'
         $name_for_self,  # no '$' here !
	 $fh,             # $fh,
	);

        $recursion_level--;
	}
      else
        {
	print $fh "\n",$spacer,'"',$key,'" => ';

        if(defined $value)
	  {
	  $value=~s/\\/\\\\/gs;
	  $value=~s/\"/\\"/gs;
	  $value=~s/\$/\\\$/gs;
	  $value=~s/\@/\\@/gs;

          print $fh '"',$value,'"';
	  }
        else
	  {
	  print $fh 'undef';
	  }
	}

      $comma=',';
      $previous_key=$key;
      } # while

    if(defined $previous_key)
      {
      print $fh "  # $access_code_marker\${$_access_strings[$recursion_level]}{$previous_key}"
      }

    print $fh "\n",$spacer;

    print $fh $anonymous_style_printing ? '}' : ')';
    }
  #++++++++++++++++++++++++
  elsif(ref $ref eq 'CODE')
  #++++++++++++++++++++++++
    {
    my $additional_spacer=' ' x $dump_indent_factor;

    print $fh 
    (0 == $recursion_level) ? '' : "\n",
    $spacer,
    ((0 == $recursion_level) and (defined $name) and ($name ne ''))
    ? "sub $name\n"
    : "sub\n",
    $spacer,
    $additional_spacer,
    join
      (
       "\n$spacer$additional_spacer",
       split("\n",$_deparse_instance->coderef2text($ref))
      )
    ;
    }
  else
    {
    die "$tool_name $error_marker unsupported reference type \$ref=$ref, died";
    }
  }

if(0 == $recursion_level)
  {
  if(not(ref($ref) eq 'CODE') or (not defined $name) or ($name eq ''))
    {
    print $fh ";";
    }

  print $fh "\n";

  unless($_dont_study)
    {
    $_ref_to_itself_number=0;
    undef(@_refs_to_itself);
    undef(%_refs_to_itself_to_numbers);
    undef($_name);
    undef($_stringified_top_level_ref);
    }
  }

pop @_access_strings;
$success_return_code;
} # sub dump_array_hash_code
#---------------------------


    ############################
sub dump_array_hash_code_to_file
    ############################
{
my
(
 $ref,
 $scope_for_name, # can be 'my' or 'local'
 $name,
 $scope_for_self, # can be 'my' or 'local'
 $name_for_self,  # no '$' here !
 $filename,
 $default_fh,
)=@_;

my $fh;

if(defined $filename and $filename ne '')
  {
  $fh=new FileHandle $filename, 'w';
  if(not defined $fh)
    {
    die "$tool_name $error_marker could not open '$filename' file for writing, died";
    }
  }
else
  {
  $fh=$default_fh;
  }

&dump_array_hash_code
(
 $ref,            # $ref,
 $scope_for_name, # $scope_for_name, # can be 'my' or 'local'
 $name,           # $name,
 $scope_for_self, # can be 'my' or 'local'
 $name_for_self,  # no '$' here !
 $fh,             # $fh, 
);

print $fh ((not (defined $name)) or ($name eq '')) ? '' : "1;\n";

if($fh ne \*STDERR and $fh ne \*STDOUT)
  {
  close($fh);
  }
} # sub dump_array_hash_code_to_file
#-----------------------------------

1;
