###############################################################################
# Copyright 2001 Sergei Steshenko
# (
#  sergei_steshenko _AT_ pop3.ru
#  sergei_steshenko _AT_ softhome.net
# ).
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the Lesser/Library GNU General Public License (LGPL) as
# published by the Free Software Foundation; either version 2 of the License,
# or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# If someone modifies the program to run with LGPL-compatible FFT, the
# he/she can also release the derived program under the terms of LGPL, as
# published by Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
###############################################################################

use FileHandle;
use Sys::Hostname;

        ###############
package ArrayHashDumper;
        ###############


use strict;

our
(
 $hash_key_sorting_and_access_code_ref,
 $dump_indent_factor,
 $dump_indent_offset,
 $recursion_level,
 $tool_name,
 $error_marker,
);

use integer;

$recursion_level=0;

my $hostname=&Sys::Hostname::hostname;#warn "\$hostname=$hostname";

my $sorted_hash_keys_prefix='___SortedHashKeys___';
my $unsorted_hash_keys_prefix='___UnsortedHashKeys___';
my $work_subdir='.ArrayHashDumperWS';


$tool_name=(reverse split('/',$0))[0];
$error_marker=':E:';


# PERL_BEGIN
#############################
# written by Sergei Steshenko
#############################

#  Copyright 2001 Sergei Steshenko 
#  (
#   sergei.steshenko@usa.net
#   sergei_steshenko@my-deja.com
#  ).
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation; either version 2 of the License, or (at your option)
#  any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
#  for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


my %recursion_levels_to_sorted_keys;

#####################################
$hash_key_sorting_and_access_code_ref=
#####################################
sub{
   my ($hash_ref)=@_;

   if(!exists $recursion_levels_to_sorted_keys{$recursion_level})
     {
     @{$recursion_levels_to_sorted_keys{$recursion_level}}=
     sort keys %{$hash_ref};
     }

   my $key=shift @{$recursion_levels_to_sorted_keys{$recursion_level}};

   if(defined $key)
     {
     return ($key,${$hash_ref}{$key},); # the returned list
     }
   else
     {
     delete $recursion_levels_to_sorted_keys{$recursion_level};
     return undef;
     }
   };
#-------------------------------------------
# PERL_END




    ###############
sub dump_array_hash
    ###############
{
my
(
 $ref, 
 $name,
 $fh,
)=@_;

if(!ref $ref)
  {
  die "$tool_name $error_marker \$ref=$ref should be a reference, died";
  }
else
  {
  my $spacer=
  ' ' x ($dump_indent_factor * $recursion_level + $dump_indent_offset);

  my $anonymous_style_printing=
  $recursion_level != 0 || !defined $name || $name eq '';

  if(ref $ref eq 'ARRAY')
    {
    if($anonymous_style_printing)
      {
      print $fh "\n",$spacer,'[';
      }
    else
      {
      print $fh $spacer,"\@$name=\n$spacer(";
      }

    my $comma='';
    my $number_of_elements=scalar @{$ref};

    my $decremented_el_number;
    my $el_number;
    ###############################################################
    for($el_number=0;$el_number < $number_of_elements;$el_number++)
    ###############################################################
      {
      $decremented_el_number=$el_number-1;

      my $element=${$ref}[$el_number];

      print $fh
      $comma,
      $decremented_el_number >= 0
      ? ' # '.$decremented_el_number.' #'
      : '';

      if(ref $element)
        {
	#################
	# recursive call:
	#################
	$recursion_level++;

        &dump_array_hash
	(
	 $element, # $ref, 
	 undef,    # $name,
	 $fh,      # $fh,
	);

        $recursion_level--;
	}
      else
        {
	print $fh "\n",$spacer;

        if(defined $element)
	  {
	  $element=~s/\\/\\\\/gs;
	  $element=~s/\"/\\"/gs;
	  $element=~s/\$/\\\$/gs;
	  $element=~s/\@/\\@/gs;

	  print $fh '"'.$element.'"';
	  }
        else
	  {
	  print $fh 'undef';
	  }
	}

      $comma=',';
      } # for($el_number=0;$el_number < $number_of_elements;$el_number++)

    if(defined $decremented_el_number)
      {
      print $fh ' # '.++$decremented_el_number.' #';
      }

    print $fh "\n",$spacer;

    print $fh $anonymous_style_printing ? ']' : ')';
    }
  elsif(ref $ref eq 'HASH')
    {
    print $fh
    $anonymous_style_printing ? "\n$spacer\{" : ($spacer,"%$name=\n$spacer(");

    my $comma='';
    my $key;
    my $value;
    ###################################################################
    while
      (
        do{
	  ($key,$value)=&{$hash_key_sorting_and_access_code_ref}($ref);
	  defined $key;
          }
      )
    ###################################################################
      {
      print $fh $comma;

      $key=~s/\\/\\\\/gs;
      $key=~s/\"/\\"/gs;
      $key=~s/\$/\\\$/gs;
      $key=~s/\@/\\@/gs;

      if(ref $value)
        {
        print $fh "\n",$spacer,'"',$key,'" => ';

	#################
	# recursive call:
	#################
	$recursion_level++;

        &dump_array_hash
	(
	 $value, # $ref, 
	 undef,  # $name,
	 $fh,    # $fh,
	);

        $recursion_level--;
	}
      else
        {


	print $fh "\n",$spacer,'"',$key,'" => ';

        if(defined $value)
	  {
	  $value=~s/\\/\\\\/gs;
	  $value=~s/\"/\\"/gs;
	  $value=~s/\$/\\\$/gs;
	  $value=~s/\@/\\@/gs;

          print $fh '"',$value,'"';
	  }
        else
	  {
	  print $fh 'undef';
	  }
	}

      $comma=',';
      } # while

    print $fh "\n",$spacer;

    print $fh $anonymous_style_printing ? '}' : ')';
    }
  else
    {
    #die "$tool_name $error_marker unsupported reference type \$ref=$ref, died";
    #warn "$tool_name $error_marker unsupported reference type \$ref=$ref";
    }
  }

if($recursion_level == 0)
  {
  print $fh ";\n";
  }
}
#-----------------------------------


###########################
sub dump_array_hash_to_file
###########################
{
my
(
 $ref,
 $name,
 $filename,
 $default_fh,
)=@_;

my $fh;

if(defined $filename and $filename ne '')
  {
  $fh=new FileHandle $filename, 'w';
  if(!defined $fh)
    {
    die "$tool_name $error_marker could not open '$filename' file for writing, died";
    }
  }
else
  {
  $fh=$default_fh;
  }

&dump_array_hash($ref,$name,$fh);

print $fh (!defined $name or $name eq '') ? '' : "1;\n";

if($fh ne \*STDERR and $fh ne \*STDOUT)
  {
  close($fh);
  }
}
#---------------------------------

1;
